﻿#ifndef WITMQTTCLIENT_H
#define WITMQTTCLIENT_H

#include <QUuid>
#include <QDir>
#include <QEventLoop>
#include <QJsonDocument>
#include <QVariantMap>
#include "tmqtt_global.h"
#include "qmqtt/qmqtt.h"

#define WIT_MQTT_TOPIC_PREFIX "WIT"
#define WIT_MQTT_TOPIC_REQ "REQ"
#define WIT_MQTT_TOPIC_REP "REP"
#define WIT_MQTT_TOPIC_CHANNEL "CHA"
#define WIT_MQTT_TOPIC_LOG "LOG"
#define WIT_MQTT_TOPIC_BUS "BUS"
#define WIT_MQTT_TOPIC_DEVICE "DEV"
#define WIT_MQTT_TOPIC_NODE "NODE"
#define WIT_MQTT_TOPIC_FILE "FILE"


struct WitMqttRpcReqMessage {
    QString uid {QUuid::createUuid().toString()}; //请求的唯一ID
    QString from; //请求来源设备
    QString to; //请求目标设备
    QString func; //请求目标函数
    QVariantMap param; //请求参数

    QVariantMap toMap() const {
        return QVariantMap {
            {"uid", this->uid},
            {"from", this->from},
            {"to", this->to},
            {"func", this->func},
            {"param", this->param},
        };
    }

    QByteArray toByteArray() const {
        return QJsonDocument::fromVariant(this->toMap()).toJson();
    }

    explicit WitMqttRpcReqMessage(){}

    explicit WitMqttRpcReqMessage(const QVariantMap &iData) {
        this->uid = iData.value("uid").toString();
        this->from = iData.value("from").toString();
        this->to = iData.value("to").toString();
        this->func = iData.value("func").toString();
        this->param = iData.value("param").toMap();
    }

    explicit WitMqttRpcReqMessage(const QByteArray &iData) {
        QVariantMap map = QJsonDocument::fromJson(iData).toVariant().toMap();
        this->uid = map.value("uid").toString();
        this->from = map.value("from").toString();
        this->to = map.value("to").toString();
        this->func = map.value("func").toString();
        this->param = map.value("param").toMap();
    }
};

struct WitMqttRpcRepMessage {
    QString uid {QUuid::createUuid().toString()}; //请求的唯一ID
    QString from; //回复的来源设备
    QString to; //回复目标设备
    QVariantMap data; //回复数据

    QVariantMap toMap() const {
        return QVariantMap {
            {"uid", this->uid},
            {"from", this->from},
            {"to", this->to},
            {"data", this->data},
        };
    }
    QByteArray toByteArray() const {
        return QJsonDocument::fromVariant(this->toMap()).toJson();
    }

    explicit WitMqttRpcRepMessage(){}

    explicit WitMqttRpcRepMessage(const QVariantMap &iData) {
        this->uid = iData.value("uid").toString();
        this->from = iData.value("from").toString();
        this->to = iData.value("to").toString();
        this->data = iData.value("data").toMap();
    }

    explicit WitMqttRpcRepMessage(const QByteArray &iData) {
        QVariantMap map = QJsonDocument::fromJson(iData).toVariant().toMap();
        this->uid = map.value("uid").toString();
        this->from = map.value("from").toString();
        this->to = map.value("to").toString();
        this->data = map.value("data").toMap();
    }
};

struct WitMqttFileMessage {
    QString uid {QUuid::createUuid().toString()}; //消息的唯一ID
    int state {0}; //状态; 1=ACK应答; 2=请求开始传输文件; 3=文件内容传输; 4=文件传输完成; 5=取消文件传输;
    QString from; //请求来源设备
    QString to; //请求目标设备
    QVariantMap data;

    QVariantMap toMap() const {
        return QVariantMap {
            {"uid", this->uid},
            {"from", this->from},
            {"to", this->to},
            {"state", this->state},
            {"data", this->data},
        };
    }

    QByteArray toByteArray() const {
        return QJsonDocument::fromVariant(this->toMap()).toJson();
    }

    explicit WitMqttFileMessage(){}

    explicit WitMqttFileMessage(const QVariantMap &iData) {
        this->uid = iData.value("uid").toString();
        this->from = iData.value("from").toString();
        this->to = iData.value("to").toString();
        this->state = iData.value("state").toInt();
        this->data = iData.value("data").toMap();
    }

    explicit WitMqttFileMessage(const QByteArray &iData) {
        QVariantMap map = QJsonDocument::fromJson(iData).toVariant().toMap();
        this->uid = map.value("uid").toString();
        this->from = map.value("from").toString();
        this->to = map.value("to").toString();
        this->state = map.value("state").toInt();
        this->data = map.value("data").toMap();
    }
};

class WitMqttRpcEventLoopPrivate;
class WitMqttRpcEventLoop : public QEventLoop
{
    Q_OBJECT
public:
    explicit WitMqttRpcEventLoop(const QString &iUid, int iTimeout, QObject *parent = nullptr);
    ~WitMqttRpcEventLoop();

signals:

public slots:
    void quit(const QString &iUid, const QByteArray &iData);
    void setUid(const QString &iUid);
    QString uid() const;
    QByteArray data() const;
    void exec(ProcessEventsFlags flags = AllEvents);

private slots:
    void onTimeout();

protected:
    const QScopedPointer<WitMqttRpcEventLoopPrivate> d_ptr;

private:
    Q_DISABLE_COPY(WitMqttRpcEventLoop)
    Q_DECLARE_PRIVATE(WitMqttRpcEventLoop)
};


class WitMqttClientPrivate;
class TMQTTSHARED_EXPORT WitMqttClient : public QMQTT::Client
{
    Q_OBJECT
public:
    explicit WitMqttClient(const QString &iHost, qint16 iPort=1883, QObject *parent = nullptr);
    ~WitMqttClient();

signals:
    void rpcRequestReceived(const WitMqttRpcReqMessage &iReq);
    void rpcResponseReceived(const QString &iUid, const QByteArray &iPayload);
    void busMessageReceived(const QMQTT::Message &iMsg);
    void fileAckReplied(const QString &iUid, const QByteArray &iPayload);
    void fileReceiveFinished(const QString &iTmpFile, const QString &iTarget);
    void fileReceiveProgress(const QVariantMap &iData);
    void fileSendProgress(const QVariantMap &iData);

public slots:
    QVariantMap rpcRequest(const QString &iTarget, const QString &iFunc, const QVariantMap &iParam, int iTimeout=120000);
    void sendRpcReply(const WitMqttRpcRepMessage &iMsg);
    bool sendFile(const QString &iTargetClientId, const QString &iSrcFilePath, const QString &iTargetFileName);
    void setFileTempDir(const QString &iPath);
    QString fileTempDir() const;
    void setFileTransPackageSize(quint64 iSize);
    quint64 fileTransPackageSize() const;
    void setFileTransTimeout(int iTimeout);
    int fileTransTimeout() const;
    void cancelFileTrans();

private slots:
    void onConnected();
    void onSubscribed(const QString &iTopic, quint8 iQos);
    void onReceived(const QMQTT::Message &iMsg);

protected:
    const QScopedPointer<WitMqttClientPrivate> d_ptr;

private:
    Q_DISABLE_COPY(WitMqttClient)
    Q_DECLARE_PRIVATE(WitMqttClient)
};


#endif // WITMQTTCLIENT_H
